# IBM DDL

## Overview

IBM PowerAI Distributed Deep Learning (or DDL) is a MPI-based
communication library, which is specifically optimized for Deep Learning
training.  An application integrated with DDL becomes an MPI-application,
which will allow the use of the `ddlrun` command to invoke the job in
parallel across a cluster of systems. DDL understands multi-tier network
environment and uses different libraries (e.g. NCCL) and algorithms to
get the best performance in multi-node, multi-GPU environments.

## Required Libraries

Prerequisite packages required for IBM PowerAI Distributed Deep
Learning are provided with PowerAI:

   1. Spectrum MPI with CUDA Support
   1. NVIDIA NCCL

## ddlrun

DDL enabled programs should be launched using the `ddlrun` tool. This tool performs the following tasks automatically:

* Determines the necessary arguments to pass to MPI based on the current environment and version of MPI.
* Tests connections to the hosts, including the correct setup of ssh keys.
* Verifies that DDL is installed across the hosts.
* Verifies that the hosts all have the same hardware configuration.
* Detects the hosts' hardware configuration, including GPU count, and generates a valid topology.
* Generates the necessary rankfile, providing options to specify more specific topology details.
* Constructs, displays and executes the `mpirun` command needed to distribute  jobs to each node.

*Note: Use of the `--rankfile` option to pass in a custom rank file bypasses much of the automatic error checking, including connection tests and key validation.*

Various options can be passed to `ddlrun` to control how and where the processes are executed.

`Usage: ddlrun [options] <program> [<args>]`

### Options

```text
  <program>         The program executable. This is
                    identified as the first non-recognized
                    argument to ddlrun.

  <args>            Pass these run-time arguments to every
                    instance of <program> started by mpirun.

  -m,--m,-mode,--mode <arg0>
                    Specify <arg0> as the ddl mode to use.
                    Default is 'p'.

  -ddlarg,--ddlarg,-ddloption,--ddloption <arg0>
                    Pass <arg0> through as an argument
                    to ddl.

  -no_ddloptions,--no_ddloptions
                    Do not pass the ddloptions env var to mpi.

  -rf,--rf,-rankfile,--rankfile <arg0>
                    Provide an MPI rank file. (Skips most error
                    checking, including connection testing and
                    key validation)

  -H,--H,-host,--host <host1,host2,...,hostN>
                    List of hosts to use when generating a rankfile.

  -hostfile,--hostfile <arg0>
                    File containing list of hosts to use when
                    generating a rankfile, with one host on each line.

  -aisles,--aisles <arg0>
                    Number of aisles. If no option given 1 is assumed.

  -racks,--racks <arg0>
                    Number of racks in each aisle. If no option given 1 is
                    assumed.

  -nodes,--nodes <arg0>
                    Number of nodes per rack. If no option given the
                    number of hosts given is assumed.

  -accelerators,--accelerators <arg0>
                    Number of accelerators per node. If no option given
                    the number in the current machine is assumed. Error on
                    values larger than 16.

  -sockets,--sockets <arg0>
                    Number of CPU sockets per node. Values can be 1 or any
                    even number 2 or larger. If no option given the number
                    in the current machine is assumed.

  -cores,--cores <arg0>
                    Number of cores per socket. If no option given the
                    number in the current machine is assumed.

  -all_cores,--all_cores
                    Each instance can use all cores. Default is to equally
                    divide cores between instances.

  -mpiarg,--mpiarg <arg0>
                    Pass <arg0> through as an argument
                    to the mpirun command.

  -tcp,-TCP,--tcp,--TCP
                    Have mpi use TCP instead of infiniband to
                    communicate between nodes.

  -v--v,-verbose,--verbose
                    Provide more verbose output.

  -s,--s,-skipchecks,--skipchecks
                    Skip all diagnostic checks.

  -h,--h,-help,--help
                    display this help and exit
```

### Example Usage

The following example uses `ddlrun` to launch the `mnist-env.py` script across 2 nodes named `host1` and `host2`.
Please make sure there is a conda environment activated with the `ddl-tensorflow` package installed before running the following command.

```bash
>$ ddlrun -H host1,host2 python mnist-env.py
```

### DDL Modes

DDL can perform an allreduce using different algorithms. These are specified with the mode argument to `ddlrun`. If no mode is selected, `ddlrun` will select `p` mode. Note, different modes can be used between different subsections of a cluster. For example, `n` mode can be used between GPUs within the same node, then `r` mode can be used between GPUs in different nodes. To specify this, use mode `nr`. Below are the different modes that DDL supports.

* `b` mode : Use enhanced NCCL allreduce algorithm.

* `n` mode : Use NCCL allreduce algorithm.

* `r` mode : Use MPI-based ring algorithm.

* `c` mode : Use an MPI-based recursive algorithm. This will only work when the number of nodes is a power of 2.

* `m` mode : Use MPI's allreduce algorithm. This mode can work on machines without GPUs.

* `p` mode : Automatically picks the best mode to use.

### DDL options

There are a number of runtime options for the DDL engine. These options can be passed to `ddlrun` using the `--ddloption` flag.
*  -mpi_mode <N> : Allows to turn on or off MPI based communication. By default, the value is 1.

* `-cross_node_nccl <N>`: Allow cross-node communication to occur using NCCL as the backend. When `N` is non-zero, NCCL can be used for cross-node communication. By default, the value is `1.

* `-pipe <N>`: Run DDL in N staged pipeline. The default is 0 (which means off). 1 has no performance benefit. 2 can sometimes see a performance benefit. Note: the `-pipe` option does not work with modes: `r`, `c` or `m`.

* `-pack <F>`: Use 16-bit floating point numbers when `F` is non-zero. `F` is a scaling factor. For example, if `F` is `2.0`, it will scale up x2 before converting to 16-bit floating point numbers, then x0.5 before converting back to 32-bit floating point numbers.

* `-cpu_only <N>`: When `N` is non-zero, DDL allreduce operations will only work with arrays on the CPU and will only operate on the CPU. By default, the value is `0`. NOTE: This option will only work if running on a machine with CUDA installed.

* `-dbg_level <0,1,2>`: 0 for no, 1 for mild, and 2 for detailed debug messages

* `-dump_iter <N>`: Outputs DDL performance information every `N` iterations.

When `dump_iter` is given, you can see the following periodically where
you can find which learner has the maximum jitter and end to end DDL
elapsed time. Also, per dimension, it shows runtime breakdown along with
the selected algorithm for that particular dimension.

![Alt text](ddl_dump.png?raw=true "DDL dump")

## Integration with Deep Learning Frameworks

IBM PowerAI Distributed Deep Learning has been integrated with the
PowerAI IBM Caffe and TensorFlow packages. `ddlrun` should be used to
launch training using the IBM PowerAI Distributed Deep Learning
integration.

* Caffe

    IBM PowerAI Distributed Deep Learning is directly integrated into Caffe, and can be exercised by using the `ddlrun` to launch caffe.

* Pytorch

    IBM PowerAI Distributed Deep Learning is directly integrated into Pytorch, in the form of `ddl` backend in Pytorch's communication package `torch.distributed`.

    The PowerAI Pytorch package provides example training setup based on Pytorch's training model scripts. These can be found on your system in:

    ```text
    $CONDA_PREFIX/lib/python$PY_VER/site-packages/torch/examples/ddl_examples/
    ```

    More details on IBM PowerAI Distributed Deep Learning integration into Pytorch, can be found in:

    ```text
    $CONDA_PREFIX/doc/ddl/DDL-Pytorch-Tutorial.md
    ```

* TensorFlow

    DDL is indirectly integrated into TensorFlow in the form of a custom
    operator. The custom operator is provided as a shared library, which is
    loaded and invoked in the python training script.

    The PowerAI ddl-tensorflow package provides an example training
    setup based on the TensorFlow High Performance models from the
    TensorFlow benchmarks repository and the TensorFlow-Slim model library from
    the TensorFlow models repository. Those can be found on your system in:

    ```text
    $CONDA_PREFIX/ddl-tensorflow/examples/
    ```

    More details on IBM PowerAI Distributed Deep Learning integration into TensorFlow, can be found in:

    ```text
    $CONDA_PREFIX/doc/ddl-tensorflow/README.md
    ```
---

## DDL API

The IBM Distributed Deep Learning (DDL) library is provided as a C library. For more information on how to use DDL from within a C/C++ program, see `$CONDA_PREFIX/doc/ddl/DDL-API-Examples.md`.

A Python wrapper for the DDL library is also provided. For more information on how to use DDL from within a Python script, see `$CONDA_PREFIX/doc/ddl/DDL-API-Python-Examples.md`.

## Tips

### Non-Default Spectrum

To use a version of IBM Spectrum MPI that is installed in a non-default location,
the environment variable `MPI_ROOT` must be set on the machine from which the
`mpirun` command is executed. `MPI_ROOT` must point to the directory which contains
the `bin` and `lib` directories. For example, if the user wishes to use the version
of IBM Spectrum MPI that is installed in `/opt/ibm/my_spectrum_mpi` with the binaries
in `/opt/ibm/my_spectrum_mpi/bin`, then they must set
`export MPI_ROOT="/opt/ibm/my_spectrum_mpi"`.

### Using `localhost` with Spectrum MPI

Using `localhost` as a host when using Spectrum MPI can cause an unintuitive error. To run a ddl program on the local machine either do not specify a host list, if only one host is being used, or specify an actual IP address or host name instead of `localhost`.

### `ddlrun` hangs with --rankfile option

Use of the `--rankfile` option to pass in a custom rank file to `ddlrun` bypasses much of the automatic error checking, including connection tests and key validation. This can lead to hangs if ssh keys are not properly configured on all hosts.

### `ddlrun` hanging on hosts with multiple interfaces

It is possible in a distributed environment to have more than one
interface for each host. In such a scenario, OpenMPI by default, uses
any and all interfaces that are "up" to communicate with a host. To
avoid problems in such cases you can tell MPI to use given
interface. E.g.:

```bash
$ ddlrun --mpiarg "--mca btl_tcp_if_include ib0" ...

$ ddlrun --mpiarg "--mca btl_tcp_if_exclude lo,enp1s0f2" ...
```

More details available on the [Spectrum MPI website](https://www.ibm.com/support/knowledgecenter/en/SSZTET_10.1.0/smpi02/smpi02_setting_mca_mpirun.html).

### Unable to install Mellanox OFED within a container

Some versions of the Mellanox OFED (MOFED) installer contain a bug preventing installation with the "user-space" flag within a container. MOFED version 4.3-3 can be used to work around this problem.

### Using Ubuntu as the host OS

Spectrum MPI requires the usage of ptrace. By default, Ubuntu does not allow this. In order to turn on this capability, the following command should be run on the host system:

```bash
$ sudo bash -c "echo '0' > /proc/sys/kernel/yama/ptrace_scope"
```

---

&copy; Copyright IBM Corporation 2018, 2019
