"""
*****************************************************************

Licensed Materials - Property of IBM

(C) Copyright IBM Corp. 2018, 2019. All Rights Reserved.

US Government Users Restricted Rights - Use, duplication or
disclosure restricted by GSA ADP Schedule Contract with IBM Corp.

*****************************************************************
"""

import ddlrun.error as error
import ddlrun.utilities as utilities


def verify_host_configs(args):
    """Use the rank_gen program to verify that all of the hosts have the same
        hardware configuration.

    Args:
        args: Parsed program arguments.

    Returns:
        Smallest number of cores on a host.

    Raises:
        GenericDDLError
        DDLError
    """
    host_configs = list()
    host_cores = list()
    encountered_error = False
    for host in args.hosts:
        ret_code, host_config, stderr = utilities.mpirun_command(
            "rank_gen.py --machine_report %s %s" % (utilities.make_arg_list(
                args.hw_args), host),
            host,
            conda_env=args.conda_env,
            args=args)
        error.log_info(host_config)
        error.log_info(stderr)
        if ret_code:
            split_config = host_config.split(",")
            host_cores.append(split_config[1])
            del split_config[1]
            host_configs.append(",".join(split_config))
        else:
            encountered_error = True
            error.log_error(error.ErrorMessage.HOST_CFG, host)
            continue

    if encountered_error:
        raise error.DDLError()

    # Check to see if all of the configs are the same by converting the list
    # to a set and counting the entries.
    if not host_configs or len(set(host_configs)) > 1:
        for i in range(len(args.hosts)):
            split_config = host_configs[i].split(",")
            error.log_info("Host: %s\tAccelerators: %s\tSockets: %s" %
                           (args.hosts[i], split_config[0], split_config[1]))
        raise error.GenericDDLError(error.ErrorMessage.DIFF_HOST_CFG)

    if not host_cores or len(set(host_cores)) > 1:
        error.log_warning(error.ErrorMessage.DIFF_CORES)

    smallest_cores = int(host_cores[0])
    for cores in host_cores:
        if int(cores) < smallest_cores:
            smallest_cores = int(cores)

    return smallest_cores


def verify_infiniband_config(args):
    """Verify that the MOFED user-space packages are installed by running the ofed_info command
        on all the hosts when using Spectrum MPI and InfiniBand.

    Args:
        args: Parsed program arguments.

    Returns:
        None

    Raises:
        DDLError
    """
    # Only check for Spectrum MPI and when -tcp and -pami_noib are not used
    if not args.use_spectrum or "-tcp" in args.mpi_args or "-pami_noib" in args.mpi_args:
        return

    encountered_error = False
    for host in args.hosts:
        ret_code, _, _ = utilities.mpirun_command(
            "ofed_info -n", host, conda_env=args.conda_env, args=args)
        if not ret_code:
            encountered_error = True
            error.log_error(error.ErrorMessage.INFINIBAND_NOT_SETUP, host)

    if encountered_error:
        raise error.DDLError()
